@@ -124,6 +124,7 @@ class BeatmapsetEventType(EnumModel):
     NOMINATE = "nominate"
     NOMINATE_MODES = "nominate_modes"
     NOMINATION_RESET = "nomination_reset"
+    NOMINATION_RESET_RECEIVED = "nomination_reset_received"
     QUALIFY = "qualify"
     RANK = "rank"
     REMOVE_FROM_LOVED = "remove_from_loved"
@@ -150,9 +151,8 @@ class UserListViews(EnumModel):
     BRICK = "brick"
 
 class KudosuAction(EnumModel):
-    # TODO ideally these wouldn't be prefixed with ``vote``. They aren't
-    # documented as such in https://osu.ppy.sh/docs/index.html#kudosuhistory,
-    # but that's what the api returns
+    # names don't match documentation in
+    # https://osu.ppy.sh/docs/index.html#kudosuhistory
     GIVE = "vote.give"
     RESET = "vote.reset"
     REVOKE = "vote.revoke"
@@ -244,6 +244,22 @@ class BeatmapDiscussionPostSort(EnumModel):
     NEW = "id_desc"
     OLD = "id_asc"
 
+class BeatmapsetStatus(EnumModel):
+    ALL = "all"
+    RANKED = "ranked"
+    QUALIFIED = "qualified"
+    DISQUALIFIED = "disqualified"
+    NEVER_QUALIFIED = "never_qualified"
+
+class ChannelType(EnumModel):
+    PUBLIC = "PUBLIC"
+    PRIVATE = "PRIVATE"
+    MULTIPLAYER = "MULTIPLAYER"
+    SPECTATOR = "SPECTATOR"
+    TEMPORARY = "TEMPORARY"
+    PM = "PM"
+    GROUP = "GROUP"
+
 class CommentableType(EnumModel):
     NEWS_POST = "news_post"
     CHANGELOG = "build"
@@ -465,6 +481,10 @@ class ForumPostBody(Model):
     html: str
     raw: str
 
+@dataclass
+class ReviewsConfig(Model):
+    max_blocks: int
+
 
 # ===================
 # Undocumented Models
