@@ -14,8 +14,9 @@ from ossapi.enums import (UserAccountHistory, ProfileBanner, UserBadge, Country,
     KudosuGiver, KudosuPost, EventType, EventAchivement, EventUser,
     EventBeatmap, BeatmapsetApproval, EventBeatmapset, KudosuVote,
     BeatmapsetEventType, UserRelationType, UserLevel, UserGradeCounts,
-    GithubUser, ChangelogSearch, ForumTopicType, ForumPostBody, ForumTopicSort)
-from ossapi.utils import Datetime, Model
+    GithubUser, ChangelogSearch, ForumTopicType, ForumPostBody, ForumTopicSort,
+    ChannelType, ReviewsConfig)
+from ossapi.utils import Datetime, Model, Expandable
 
 T = TypeVar("T")
 S = TypeVar("S")
@@ -35,7 +36,7 @@ is, not that the api actually lets any type be returned there.
 # =================
 
 @dataclass
-class UserCompact(Model):
+class UserCompact(Model, Expandable):
     """
     https://osu.ppy.sh/docs/index.html#usercompact
     """
@@ -52,8 +53,7 @@ class UserCompact(Model):
     is_supporter: bool
     last_visit: Optional[Datetime]
     pm_friends_only: bool
-    # TODO pretty sure this needs to be optional but it's not documented as
-    # such, open an issue?
+    # documented as non-optional
     profile_colour: Optional[str]
     username: str
 
@@ -108,6 +108,9 @@ class UserCompact(Model):
     # deprecated, replaced by rank_history
     rankHistory: Optional[RankHistory]
 
+    def expand(self):
+        return self._api.user(self)
+
 @dataclass
 class User(UserCompact):
     comments_count: int
@@ -130,10 +133,13 @@ class User(UserCompact):
     twitter: Optional[str]
     website: Optional[str]
 
+    def expand(self):
+        # we're already expanded, no need to waste an api call
+        return self
 
 
 @dataclass
-class BeatmapCompact(Model):
+class BeatmapCompact(Model, Expandable):
     # required fields
     # ---------------
     difficulty_rating: float
@@ -147,10 +153,13 @@ class BeatmapCompact(Model):
     # optional fields
     # ---------------
     beatmapset: Optional[BeatmapsetCompact]
+    beatmapset_id: Optional[int]
     checksum: Optional[str]
     failtimes: Optional[Failtimes]
     max_combo: Optional[int]
 
+    def expand(self):
+        return self._api.beatmap(self)
 
 @dataclass
 class Beatmap(BeatmapCompact):
@@ -182,7 +191,10 @@ class Beatmap(BeatmapCompact):
     # -----------------
     beatmapset: Optional[Beatmapset]
 
+    def expand(self):
+        return self
 
+# TODO make this expandable when we implement beatmapset lookup
 @dataclass
 class BeatmapsetCompact(Model):
     """
@@ -226,6 +238,8 @@ class BeatmapsetCompact(Model):
     recent_favourites: Optional[Any]
     related_users: Optional[Any]
     user: Optional[UserCompact]
+    # undocumented
+    track_id: Optional[int]
 
 @dataclass
 class Beatmapset(BeatmapsetCompact):
@@ -265,7 +279,7 @@ class Score(Model):
     perfect: bool
     statistics: Statistics
     # documented as non-optional in docs but broken beatmaps like acid rain
-    # (1981090) have scores with null pp values. TODO open issue
+    # (1981090) have scores with null pp values
     pp: Optional[float]
     rank: Grade
     created_at: Datetime
@@ -484,6 +498,9 @@ class BeatmapsetDiscussion(Model):
     can_be_resolved: bool
     can_grant_kudosu: bool
     created_at: Datetime
+    # documented as non-optional, but just try api.beatmapsets_events() to see
+    # it can't be non-optional
+    current_user_attributes: Optional[Any]
     updated_at: Datetime
     deleted_at: Optional[Datetime]
     # documented as non-optional
@@ -520,8 +537,7 @@ class KudosuHistory(Model):
     created_at: Datetime
     giver: Optional[KudosuGiver]
     post: KudosuPost
-    # TODO type hint properly when https://github.com/ppy/osu-web/issues/7549 is
-    # resolved
+    # see https://github.com/ppy/osu-web/issues/7549
     details: Any
 
 @dataclass
@@ -714,7 +730,7 @@ class MultiplayerScoresCursor(Model):
 # ===================
 
 @dataclass
-class BeatmapSearchResult(Model):
+class BeatmapsetSearchResult(Model):
     beatmapsets: List[Beatmapset]
     cursor: CursorT
     recommended_difficulty: Optional[float]
@@ -722,6 +738,15 @@ class BeatmapSearchResult(Model):
     total: int
     search: Any
 
+@dataclass
+class BeatmapsetDiscussions(Model):
+    beatmaps: List[Beatmap]
+    cursor: CursorT
+    discussions: List[BeatmapsetDiscussion]
+    included_discussions: List[BeatmapsetDiscussion]
+    reviews_config: ReviewsConfig
+    users: List[UserCompact]
+
 @dataclass
 class BeatmapsetDiscussionReview(Model):
     # https://github.com/ppy/osu-web/blob/master/app/Libraries/BeatmapsetDiscussionReview.php
@@ -787,13 +812,18 @@ class BeatmapsetEventCommentOwnerChange(BeatmapsetEventCommentNone):
 class BeatmapsetEventCommentNominate(Model):
     # for some reason this comment type doesn't have the normal
     # beatmap_discussion_id and beatmap_discussion_post_id attributes (they're
-    # not even null, just missing). TODO Open an issue on osu-web?
+    # not even null, just missing).
     modes: List[GameMode]
 
 @dataclass
 class BeatmapsetEventCommentWithNominators(BeatmapsetEventCommentNoPost):
     nominator_ids: Optional[List[int]]
 
+@dataclass
+class BeatmapsetEventCommentWithSourceUser(BeatmapsetEventCommentNoPost):
+    source_user_id: int
+    source_user_username: str
+
 @dataclass
 class BeatmapsetEvent(Model):
     # https://github.com/ppy/osu-web/blob/master/app/Models/BeatmapsetEvent.php
@@ -811,7 +841,7 @@ class BeatmapsetEvent(Model):
         mapping = {
             BeatmapsetEventType.BEATMAP_OWNER_CHANGE: BeatmapsetEventCommentOwnerChange,
             BeatmapsetEventType.DISCUSSION_DELETE: BeatmapsetEventCommentNoPost,
-            # TODO: ``api.beatmapsets_events(types=[BeatmapsetEventType.DISCUSSION_LOCK])``
+            # ``api.beatmapsets_events(types=[BeatmapsetEventType.DISCUSSION_LOCK])``
             # doesn't seem to be recognized, just returns all events. Was this
             # type discontinued?
             # BeatmapsetEventType.DISCUSSION_LOCK: BeatmapsetEventComment,
@@ -837,6 +867,7 @@ class BeatmapsetEvent(Model):
             # same here
             # BeatmapsetEventType.NOMINATE_MODES: BeatmapsetEventComment,
             BeatmapsetEventType.NOMINATION_RESET: BeatmapsetEventCommentWithNominators,
+            BeatmapsetEventType.NOMINATION_RESET_RECEIVED: BeatmapsetEventCommentWithSourceUser,
             BeatmapsetEventType.QUALIFY: type(None),
             BeatmapsetEventType.RANK: type(None),
             BeatmapsetEventType.REMOVE_FROM_LOVED: BeatmapsetEventCommentLovedRemoval,
@@ -845,7 +876,44 @@ class BeatmapsetEvent(Model):
         type_ = BeatmapsetEventType(self.type)
         return {"comment": mapping[type_]}
 
+@dataclass
+class ChatChannel(Model):
+    channel_id: int
+    description: Optional[str]
+    icon: str
+    # documented as non-optional (try pming tillerino with this non-optional)
+    moderated: Optional[bool]
+    name: str
+    type: ChannelType
 
+    # optional fields
+    # ---------------
+    first_message_id: Optional[int]
+    last_message_id: Optional[int]
+    last_read_id: Optional[int]
+    recent_messages: Optional[List[ChatMessage]]
+    users: Optional[List[int]]
+
+@dataclass
+class ChatMessage(Model):
+    channel_id: int
+    content: str
+    is_action: bool
+    message_id: int
+    sender: UserCompact
+    sender_id: int
+    timestamp: Datetime
+
+@dataclass
+class CreatePMResponse(Model):
+    message: ChatMessage
+    new_channel_id: int
+
+    # undocumented
+    channel: ChatChannel
+
+    # documented but not present in response
+    presence: Optional[List[ChatChannel]]
 
 @dataclass
 class ModdingHistoryEventsBundle(Model):
