@@ -1,39 +1,46 @@
-from typing import (get_type_hints, get_origin, get_args, Union, TypeVar,
-    Optional, List, _GenericAlias)
+from typing import Union, TypeVar, Optional, List, _GenericAlias
 import logging
 import webbrowser
 import socket
 import pickle
 from pathlib import Path
-from tempfile import NamedTemporaryFile
 from datetime import datetime
 from enum import Enum
 from urllib.parse import unquote
 import inspect
 import json
 from keyword import iskeyword
+import hashlib
+import functools
 
 from requests_oauthlib import OAuth2Session
-from oauthlib.oauth2 import BackendApplicationClient
+from oauthlib.oauth2 import (BackendApplicationClient, TokenExpiredError,
+    AccessDeniedError)
+from oauthlib.oauth2.rfc6749.errors import InsufficientScopeError
+import osrparse
+from typing_utils import issubtype, get_type_hints, get_origin, get_args
 
-from ossapi.models import (Beatmap, BeatmapUserScore, ForumTopicAndPosts,
-    Search, CommentBundle, Cursor, Score, BeatmapSearchResult,
-    ModdingHistoryEventsBundle, User, Rankings, BeatmapScores, KudosuHistory,
-    Beatmapset, BeatmapPlaycount, Spotlight, Spotlights, WikiPage, _Event,
-    Event, BeatmapsetDiscussionPosts, Build, ChangelogListing,
-    MultiplayerScores, MultiplayerScoresCursor, BeatmapsetDiscussionVotes)
-from ossapi.mod import Mod
+from ossapi.models import (Beatmap, BeatmapCompact, BeatmapUserScore,
+    ForumTopicAndPosts, Search, CommentBundle, Cursor, Score,
+    BeatmapsetSearchResult, ModdingHistoryEventsBundle, User, Rankings,
+    BeatmapScores, KudosuHistory, Beatmapset, BeatmapPlaycount, Spotlight,
+    Spotlights, WikiPage, _Event, Event, BeatmapsetDiscussionPosts, Build,
+    ChangelogListing, MultiplayerScores, MultiplayerScoresCursor,
+    BeatmapsetDiscussionVotes, CreatePMResponse, BeatmapsetDiscussions,
+    UserCompact)
 from ossapi.enums import (GameMode, ScoreType, RankingFilter, RankingType,
     UserBeatmapType, BeatmapDiscussionPostSort, UserLookupKey,
     BeatmapsetEventType, CommentableType, CommentSort, ForumTopicSort,
     SearchMode, MultiplayerScoresSort, BeatmapsetDiscussionVote,
-    BeatmapsetDiscussionVoteSort)
+    BeatmapsetDiscussionVoteSort, BeatmapsetStatus, MessageType)
 from ossapi.utils import (is_compatible_type, is_primitive_type, is_optional,
-    is_base_model_type, is_model_type, is_high_model_type)
+    is_base_model_type, is_model_type, is_high_model_type, Expandable)
+from ossapi.mod import Mod
+from ossapi.replay import Replay
 
 # our ``request`` function below relies on the ordering of these types. The
 # base type must come first, with any auxiliary types that the base type accepts
-# cooming after.
+# coming after.
 # These types are intended to provide better type hinting for consumers. We
 # want to support the ability to pass ``"osu"`` instead of ``GameMode.STD``,
 # for instance. We automatically convert any value to its base class if the
@@ -55,100 +62,318 @@ SearchModeT = Union[SearchMode, str]
 MultiplayerScoresSortT = Union[MultiplayerScoresSort, str]
 BeatmapsetDiscussionVoteT = Union[BeatmapsetDiscussionVote, int]
 BeatmapsetDiscussionVoteSortT = Union[BeatmapsetDiscussionVoteSort, str]
+MessageTypeT = Union[MessageType, str]
+BeatmapsetStatusT = Union[BeatmapsetStatus, str]
 
-def request(function):
+BeatmapIdT = Union[int, BeatmapCompact]
+UserIdT = Union[int, UserCompact]
+
+def request(scope, *, requires_login=False):
     """
-    Automatically instantiates parameters with their type hint if the type hint
-    is a union of a base class. This means, for instance, that a function which
-    accepts a ``ModT`` will have the value of that parameter automatically
-    converted to a ``Mod``, even if the user passes a `str`.
+    Handles various validation and preparation tasks for any endpoint request
+    method.
+
+    This method does the following things:
+    * makes sure the client has the requuired scope to access the endpoint in
+      question
+    * makes sure the client has the right grant to access the endpoint in
+      question (the client credentials grant cannot access endpoints which
+      require the user to be "logged in", such as downloading a replay)
+    * converts parameters to an instance of a base model if the parameter is
+      annotated as being a base model. This means, for instance, that a function
+      with an argument annotated as ``ModT`` (``Union[Mod, str, int, list]``)
+      will have the value of that parameter automatically converted to a
+      ``Mod``, even if the user passes a `str`.
+    * converts arguments of type ``BeatmapIdT`` or ``UserIdT`` into a beatmap or
+      user id, if the passed argument was a ``BeatmapCompact`` or
+      ``UserCompact`` respectively.
+
+    Parameters
+    ----------
+    scope: Scope
+        The scope required for this endpoint. If ``None``, no scope is required
+        and any authenticated cliient can access it.
+    requires_login: bool
+        Whether this endpoint requires a "logged-in" client to retrieve it.
+        Currently, only authtorization code grants can access these endpoints.
     """
-    instantiate = {}
-    for name, type_ in function.__annotations__.items():
-        origin = get_origin(type_)
-        args = get_args(type_)
-        if origin is Union and is_base_model_type(args[0]):
-            instantiate[name] = args[0]
+    def decorator(function):
+        instantiate = {}
+        for name, type_ in function.__annotations__.items():
+            origin = get_origin(type_)
+            args = get_args(type_)
+            if origin is Union and is_base_model_type(args[0]):
+                instantiate[name] = args[0]
 
-    arg_names = list(inspect.signature(function).parameters)
+        arg_names = list(inspect.signature(function).parameters)
 
-    def wrapper(*args, **kwargs):
-        # we may need to edit this later so convert from tuple
-        args = list(args)
+        @functools.wraps(function)
+        def wrapper(*args, **kwargs):
+            self = args[0]
+            if scope is not None and scope not in self.scopes:
+                raise InsufficientScopeError(f"A scope of {scope} is required "
+                    "for this endpoint. Your client's current scopes are "
+                    f"{self.scopes}")
 
-        # args and kwargs are handled separately, but in a similar fashion.
-        # The difference is that for ``args`` we need to know the name of the
-        # argument so we can look up its type hint and see if it's a parameter
-        # we need to convert.
-        for i, (arg, arg_name) in enumerate(zip(args, arg_names)):
-            if arg_name in instantiate:
-                type_ = instantiate[arg_name]
-                args[i] = type_(arg)
+            if requires_login and self.grant is Grant.CLIENT_CREDENTIALS:
+                raise AccessDeniedError("To access this endpoint you must be "
+                    "authorized using the authorization code grant. You are "
+                    "currently authorized with the client credentials grant")
 
-        for arg in kwargs:
-            if arg in instantiate:
-                type_ = instantiate[arg]
-                kwargs[arg] = type_(kwargs[arg])
+            # we may need to edit this later so convert from tuple
+            args = list(args)
 
-        return function(*args, **kwargs)
-    return wrapper
+            def is_id_type(arg_name, arg):
+                annotations = function.__annotations__
+                if arg_name not in annotations:
+                    return False
+                arg_type = annotations[arg_name]
+                if (not issubtype(BeatmapIdT, arg_type) and not
+                    issubtype(UserIdT, arg_type)):
+                    return False
+                return isinstance(arg, (BeatmapCompact, UserCompact))
+
+            # args and kwargs are handled separately, but in a similar fashion.
+            # The difference is that for ``args`` we need to know the name of
+            # the argument so we can look up its type hint and see if it's a
+            # parameter we need to convert.
+
+            for i, (arg_name, arg) in enumerate(zip(arg_names, args)):
+                if arg_name in instantiate:
+                    type_ = instantiate[arg_name]
+                    args[i] = type_(arg)
+                if is_id_type(arg_name, arg):
+                    args[i] = arg.id
+
+            for arg_name, arg in kwargs.items():
+                if arg_name in instantiate:
+                    type_ = instantiate[arg_name]
+                    kwargs[arg_name] = type_(arg)
+                if is_id_type(arg_name, arg):
+                    kwargs[arg_name] = arg.id
+
+            return function(*args, **kwargs)
+        return wrapper
+    return decorator
+
+
+class Grant(Enum):
+    CLIENT_CREDENTIALS = "client"
+    AUTHORIZATION_CODE = "authorization"
+
+class Scope(Enum):
+    CHAT_WRITE = "chat.write"
+    DELEGATE = "delegate"
+    FORUM_WRITE = "forum.write"
+    FRIENDS_READ = "friends.read"
+    IDENTIFY = "identify"
+    PUBLIC = "public"
 
 
 class OssapiV2:
+    """
+    A wrapper around osu api v2.
+
+    Parameters
+    ----------
+    client_id: int
+        The id of the client to authenticate with.
+    client_secret: str
+        The secret of the client to authenticate with.
+    redirect_uri: str
+        The redirect uri for the client. Must be passed if using the
+        authorization code grant. This must exactly match the redirect uri on
+        the client's settings page. Additionally, in order for ossapi to receive
+        authentication from this redirect uri, it must be a port on localhost.
+        So "http://localhost:3914/", "http://localhost:727/", etc are all valid
+        redirect uris. You can change your client's redirect uri from its
+        settings page.
+    scopes: List[str]
+        What scopes to request when authenticating.
+    grant: Grant or str
+        Which oauth grant (aka flow) to use when authenticating with the api.
+        Currently the api offers the client credentials (pass "client" for this
+        parameter) and authorization code (pass "authorization" for this
+        parameter) grants.
+        |br|
+        The authorization code grant requires user interaction to authenticate
+        the first time, but grants full access to the api. In contrast, the
+        client credentials grant does not require user interaction to
+        authenticate, but only grants guest user access to the api. This means
+        you will not be able to do things like download replays on the client
+        credentials grant.
+        |br|
+        If not passed, the grant will be automatically inferred as follows: if
+        ``redirect_uri`` is passed, use the authorization code grant. If
+        ``redirect_uri`` is not passed, use the client credentials grant.
+    strict: bool
+        Whether to run in "strict" mode. In strict mode, ossapi will raise an
+        exception if the api returns an attribute in a response which we didn't
+        expect to be there. This is useful for developers which want to catch
+        new attributes as they get added. More checks may be added in the future
+        for things which developers may want to be aware of, but normal users do
+        not want to have an exception raised for.
+        |br|
+        If you are not a developer, you are very unlikely to want to use this
+        parameter.
+    token_directory: str
+        If passed, the given directory will be used to store and retrieve token
+        files instead of locally wherever ossapi is installed. Useful if you
+        want more control over token files.
+    token_key: str
+        If passed, the given key will be used to name the token file instead of
+        an automatically generated one. Note that if you pass this, you are
+        taking responsibility for making sure it is unique / unused, and also
+        for remembering the key you passed if you wish to eg remove the token in
+        the future, which requires the key.
+    """
     TOKEN_URL = "https://osu.ppy.sh/oauth/token"
     AUTH_CODE_URL = "https://osu.ppy.sh/oauth/authorize"
     BASE_URL = "https://osu.ppy.sh/api/v2"
 
-    AUTHORIZATION_TOKEN_FILE = (Path(__file__).parent /
-        "authorization_code.pickle")
+    def __init__(self,
+        client_id: int,
+        client_secret: str,
+        redirect_uri: Optional[str] = None,
+        scopes: List[Union[str, Scope]] = [Scope.PUBLIC],
+        *,
+        grant: Optional[Union[Grant, str]] = None,
+        strict: bool = False,
+        token_directory: Optional[str] = None,
+        token_key: Optional[str] = None,
+    ):
+        if not grant:
+            grant = (Grant.AUTHORIZATION_CODE if redirect_uri else
+                Grant.CLIENT_CREDENTIALS)
+        grant = Grant(grant)
 
-    def __init__(self, client_id, client_secret, redirect_uri=None,
-        scopes=["public"], strict=False):
+        self.grant = grant
+        self.client_id = client_id
+        self.client_secret = client_secret
+        self.redirect_uri = redirect_uri
+        self.scopes = [Scope(scope) for scope in scopes]
         self.strict = strict
+
         self.log = logging.getLogger(__name__)
+        self.token_key = token_key or self.gen_token_key(self.grant,
+            self.client_id, self.client_secret, self.scopes)
+        self.token_directory = (
+            Path(token_directory) if token_directory else Path(__file__).parent
+        )
+        self.token_file = self.token_directory / f"{self.token_key}.pickle"
 
-        self.session = self.authenticate(client_id, client_secret, redirect_uri,
-            scopes)
-
-    def authenticate(self, client_id, client_secret, redirect_uri, scopes):
-        # if redirect_uri is not passed, assume the user wanted to use the
-        # client credentials grant.
-        if not redirect_uri:
-            if scopes != ["public"]:
+        if self.grant is Grant.CLIENT_CREDENTIALS:
+            if self.scopes != [Scope.PUBLIC]:
                 raise ValueError(f"`scopes` must be ['public'] if the "
-                    f"client credentials grant is used. Got {scopes}")
-            return self._client_credentials_grant(client_id, client_secret)
+                    f"client credentials grant is used. Got {self.scopes}")
 
-        # used saved session if it exists
-        if self.AUTHORIZATION_TOKEN_FILE.is_file():
-            with open(self.AUTHORIZATION_TOKEN_FILE, "rb") as f:
-                token = pickle.load(f)
-            return self._auth_oauth_session(client_id, client_secret, scopes,
-                token=token)
+        if self.grant is Grant.AUTHORIZATION_CODE and not self.redirect_uri:
+            raise ValueError("`redirect_uri` must be passed if the "
+                "authorization code grant is used.")
 
-        return self._authorization_code_grant(client_id, client_secret,
-            redirect_uri, scopes)
+        self.session = self.authenticate()
 
     @staticmethod
-    def clear_authentication():
-        if OssapiV2.AUTHORIZATION_TOKEN_FILE.is_file():
-            OssapiV2.AUTHORIZATION_TOKEN_FILE.unlink()
+    def gen_token_key(grant, client_id, client_secret, scopes):
+        """
+        The unique key / hash for the given set of parameters. This is intended
+        to provide a way to allow multiple OssapiV2's to live at the same time,
+        by eg saving their tokens to different files based on their key.
 
-    def _client_credentials_grant(self, client_id, client_secret):
+        This function is also deterministic, to eg allow tokens to be reused if
+        OssapiV2 is instantiated twice with the same parameters. This avoids the
+        need to reauthenticate unless absolutely necessary.
+        """
+        grant = Grant(grant)
+        scopes = [Scope(scope) for scope in scopes]
+        m = hashlib.sha256()
+        m.update(grant.value.encode("utf-8"))
+        m.update(str(client_id).encode("utf-8"))
+        m.update(client_secret.encode("utf-8"))
+        for scope in scopes:
+            m.update(scope.value.encode("utf-8"))
+        return m.hexdigest()
+
+    @staticmethod
+    def remove_token(key, token_directory=None):
+        """
+        Removes the token file associated with the given key. If
+        ``token_directory`` is passed, looks there for the token file instead of
+        locally in ossapi's install site.
+
+        To determine the key associated with a given grant, client_id,
+        client_secret, and set of scopes, use ``gen_token_key``.
+        """
+        token_directory = (
+            Path(token_directory) if token_directory else Path(__file__).parent
+        )
+        token_file = token_directory / f"{key}.pickle"
+        token_file.unlink()
+
+    def authenticate(self):
+        """
+        Returns a valid OAuth2Session, either from a saved token file associated
+        with this OssapiV2's parameters, or from a fresh authentication if no
+        such file exists.
+        """
+        if self.token_file.exists():
+            with open(self.token_file, "rb") as f:
+                token = pickle.load(f)
+
+            if self.grant is Grant.CLIENT_CREDENTIALS:
+                return OAuth2Session(self.client_id, token=token)
+
+            if self.grant is Grant.AUTHORIZATION_CODE:
+                auto_refresh_kwargs = {
+                    "client_id": self.client_id,
+                    "client_secret": self.client_secret
+                }
+                return OAuth2Session(self.client_id, token=token,
+                    redirect_uri=self.redirect_uri,
+                    auto_refresh_url=self.TOKEN_URL,
+                    auto_refresh_kwargs=auto_refresh_kwargs,
+                    token_updater=self._save_token,
+                    scope=[scope.value for scope in self.scopes])
+
+        if self.grant is Grant.CLIENT_CREDENTIALS:
+            return self._new_client_grant(self.client_id, self.client_secret)
+
+        return self._new_authorization_grant(self.client_id, self.client_secret,
+            self.redirect_uri, self.scopes)
+
+    def _new_client_grant(self, client_id, client_secret):
+        """
+        Authenticates with the api from scratch on the client grant.
+        """
         self.log.info("initializing client credentials grant")
         client = BackendApplicationClient(client_id=client_id, scope=["public"])
-        oauth = OAuth2Session(client=client)
-        oauth.fetch_token(token_url=self.TOKEN_URL,
+        session = OAuth2Session(client=client)
+        token = session.fetch_token(token_url=self.TOKEN_URL,
             client_id=client_id, client_secret=client_secret)
 
-        return oauth
+        self._save_token(token)
+        return session
 
-    def _authorization_code_grant(self, client_id, client_secret, redirect_uri,
+    def _new_authorization_grant(self, client_id, client_secret, redirect_uri,
         scopes):
+        """
+        Authenticates with the api from scratch on the authorization code grant.
+        """
         self.log.info("initializing authorization code")
-        oauth = self._auth_oauth_session(client_id, client_secret, scopes,
-            redirect_uri=redirect_uri)
-        authorization_url, _state = oauth.authorization_url(self.AUTH_CODE_URL)
+
+        auto_refresh_kwargs = {
+            "client_id": client_id,
+            "client_secret": client_secret
+        }
+        session = OAuth2Session(client_id, redirect_uri=redirect_uri,
+            auto_refresh_url=self.TOKEN_URL,
+            auto_refresh_kwargs=auto_refresh_kwargs,
+            token_updater=self._save_token,
+            scope=[scope.value for scope in scopes])
+
+        authorization_url, _state = (
+            session.authorization_url(self.AUTH_CODE_URL)
+        )
         webbrowser.open(authorization_url)
 
         # open up a temporary socket so we can receive the GET request to the
@@ -172,33 +397,41 @@ class OssapiV2:
         serversocket.close()
 
         code = data.split("code=")[1].split("&state=")[0]
-        token = oauth.fetch_token(self.TOKEN_URL, client_id=client_id,
+        token = session.fetch_token(self.TOKEN_URL, client_id=client_id,
             client_secret=client_secret, code=code)
         self._save_token(token)
 
-        return oauth
-
-    def _auth_oauth_session(self, client_id, client_secret, scopes, *,
-        token=None, redirect_uri=None):
-        auto_refresh_kwargs = {
-            "client_id": client_id,
-            "client_secret": client_secret
-        }
-        return OAuth2Session(client_id, token=token, redirect_uri=redirect_uri,
-            auto_refresh_url=self.TOKEN_URL,
-            auto_refresh_kwargs=auto_refresh_kwargs,
-            token_updater=self._save_token,
-            scope=scopes)
+        return session
 
     def _save_token(self, token):
-        self.log.info(f"saving token to {self.AUTHORIZATION_TOKEN_FILE}")
-        with open(self.AUTHORIZATION_TOKEN_FILE, "wb+") as f:
+        """
+        Saves the token to this OssapiV2's associated token file.
+        """
+        self.log.info(f"saving token to {self.token_file}")
+        with open(self.token_file, "wb+") as f:
             pickle.dump(token, f)
 
-    def _get(self, type_, url, params={}):
+    def _request(self, type_, method, url, params={}, data={}):
         params = self._format_params(params)
-        r = self.session.get(f"{self.BASE_URL}{url}", params=params)
-        self.log.info(f"made request to {r.request.url}")
+        try:
+            r = self.session.request(method, f"{self.BASE_URL}{url}",
+                params=params, data=data)
+        except TokenExpiredError:
+            # provide "auto refreshing" for client credentials grant. The client
+            # grant doesn't actually provide a refresh token, so we can't hook
+            # onto OAuth2Session's auto_refresh functionality like we do for the
+            # authorization code grant. But we can do something effectively
+            # equivalent: whenever we make a request with an expired client
+            # grant token, just request a new one.
+            if self.grant is not Grant.CLIENT_CREDENTIALS:
+                raise
+            self.session = self._new_client_grant(self.client_id,
+                self.client_secret)
+            # redo the request now that we have a valid token
+            r = self.session.request(method, f"{self.BASE_URL}{url}",
+                params=params, data=data)
+
+        self.log.info(f"made {method} request to {r.request.url}")
         json_ = r.json()
         self.log.debug(f"received json: \n{json.dumps(json_, indent=4)}")
         # TODO this should just be ``if "error" in json``, but for some reason
@@ -206,9 +439,15 @@ class OssapiV2:
         # open an issue on osu-web?
         if len(json_) == 1 and "error" in json_:
             raise ValueError(f"api returned an error of `{json_['error']}` for "
-                f"a request to {unquote(r.request.url)}")
+                f"a request to {unquote(url)}")
         return self._instantiate_type(type_, json_)
 
+    def _get(self, type_, url, params={}):
+        return self._request(type_, "GET", url, params=params)
+
+    def _post(self, type_, url, data={}):
+        return self._request(type_, "POST", url, data=data)
+
     def _format_params(self, params):
         for key, value in params.copy().items():
             if isinstance(value, list):
@@ -455,6 +694,17 @@ class OssapiV2:
                             f"{type_}")
                     self.log.info(f"ignoring unexpected parameter `{k}` from "
                         f"api response for type {type_}")
+
+        # "expandable" models need an api instance to be injected into them on
+        # instantiation. There isn't really a clean way to do this
+        # unfortunately. If the parameter name on the ``Expandable`` class ever
+        # changes from ``_api``, this will also need to be changed. And it's not
+        # very transparent as to where this parameter is coming from, from
+        # Expandable's perspective. But such is the way of life when abusing
+        # python type hints.
+        if isinstance(type_, type) and issubclass(type_, Expandable):
+            kwargs_["_api"] = self
+
         try:
             val = type_(**kwargs_)
         except TypeError as e:
@@ -472,22 +722,10 @@ class OssapiV2:
     # /beatmaps
     # ---------
 
-    @request
-    def beatmap_lookup(self,
-        checksum: Optional[str] = None,
-        filename: Optional[str] = None,
-        beatmap_id: Optional[int] = None
-    ) -> Beatmap:
-        """
-        https://osu.ppy.sh/docs/index.html#lookup-beatmap
-        """
-        params = {"checksum": checksum, "filename": filename, "id": beatmap_id}
-        return self._get(Beatmap, "/beatmaps/lookup", params)
-
-    @request
+    @request(Scope.PUBLIC)
     def beatmap_user_score(self,
-        beatmap_id: int,
-        user_id: int,
+        beatmap_id: BeatmapIdT,
+        user_id: UserIdT,
         mode: Optional[GameModeT] = None,
         mods: Optional[ModT] = None
     ) -> BeatmapUserScore:
@@ -498,9 +736,9 @@ class OssapiV2:
         return self._get(BeatmapUserScore,
             f"/beatmaps/{beatmap_id}/scores/users/{user_id}", params)
 
-    @request
+    @request(Scope.PUBLIC)
     def beatmap_scores(self,
-        beatmap_id: int,
+        beatmap_id: BeatmapIdT,
         mode: Optional[GameModeT] = None,
         mods: Optional[ModT] = None,
         type_: Optional[RankingTypeT] = None
@@ -512,38 +750,45 @@ class OssapiV2:
         return self._get(BeatmapScores, f"/beatmaps/{beatmap_id}/scores",
             params)
 
-    @request
+    @request(Scope.PUBLIC)
     def beatmap(self,
-        beatmap_id: int
+        beatmap_id: Optional[BeatmapIdT] = None,
+        checksum: Optional[str] = None,
+        filename: Optional[str] = None,
     ) -> Beatmap:
         """
-        https://osu.ppy.sh/docs/index.html#get-beatmap
+        combines https://osu.ppy.sh/docs/index.html#get-beatmap and
+        https://osu.ppy.sh/docs/index.html#lookup-beatmap
         """
-        return self._get(Beatmap, f"/beatmaps/{beatmap_id}")
+        if not (beatmap_id or checksum or filename):
+            raise ValueError("at least one of beatmap_id, checksum, or "
+                "filename must be passed")
+        params = {"checksum": checksum, "filename": filename, "id": beatmap_id}
+        return self._get(Beatmap, "/beatmaps/lookup", params)
 
 
     # /beatmapsets
     # ------------
 
-    @request
+    @request(Scope.PUBLIC)
     def beatmapset_discussion_posts(self,
-        beatmapset_session_id: Optional[int] = None,
+        beatmapset_discussion_id: Optional[int] = None,
         limit: Optional[int] = None,
         page: Optional[int] = None,
         sort: Optional[BeatmapDiscussionPostSortT] = None,
-        user_id: Optional[int] = None,
+        user_id: Optional[UserIdT] = None,
         with_deleted: Optional[bool] = None
     ) -> BeatmapsetDiscussionPosts:
         """
         https://osu.ppy.sh/docs/index.html#get-beatmapset-discussion-posts
         """
-        params = {"beatmapset_session_id": beatmapset_session_id,
+        params = {"beatmapset_discussion_id": beatmapset_discussion_id,
             "limit": limit, "page": page, "sort": sort, "user": user_id,
             "with_deleted": with_deleted}
         return self._get(BeatmapsetDiscussionPosts,
             "/beatmapsets/discussions/posts", params)
 
-    @request
+    @request(Scope.PUBLIC)
     def beatmapset_discussion_votes(self,
         beatmapset_discussion_id: Optional[int] = None,
         limit: Optional[int] = None,
@@ -551,7 +796,7 @@ class OssapiV2:
         receiver_id: Optional[int] = None,
         vote: Optional[BeatmapsetDiscussionVoteT] = None,
         sort: Optional[BeatmapsetDiscussionVoteSortT] = None,
-        user_id: Optional[int] = None,
+        user_id: Optional[UserIdT] = None,
         with_deleted: Optional[bool] = None
     ) -> BeatmapsetDiscussionVotes:
         """
@@ -564,11 +809,34 @@ class OssapiV2:
         return self._get(BeatmapsetDiscussionVotes,
             "/beatmapsets/discussions/votes", params)
 
+    @request(Scope.PUBLIC)
+    def beatmapset_discussions(self,
+        beatmapset_id: Optional[int] = None,
+        beatmap_id: Optional[BeatmapIdT] = None,
+        beatmapset_status: Optional[BeatmapsetStatusT] = None,
+        limit: Optional[int] = None,
+        message_types: Optional[List[MessageTypeT]] = None,
+        only_unresolved: Optional[bool] = None,
+        page: Optional[int] = None,
+        sort: Optional[BeatmapDiscussionPostSortT] = None,
+        user_id: Optional[UserIdT] = None,
+        with_deleted: Optional[bool] = None,
+    ) -> BeatmapsetDiscussions:
+        """
+        https://osu.ppy.sh/docs/index.html#get-beatmapset-discussions
+        """
+        params = {"beatmapset_id": beatmapset_id, "beatmap_id": beatmap_id,
+            "beatmapset_status": beatmapset_status, "limit": limit,
+            "message_types": message_types, "only_unresolved": only_unresolved,
+            "page": page, "sort": sort, "user": user_id,
+            "with_deleted": with_deleted}
+        return self._get(BeatmapsetDiscussions,
+            "/beatmapsets/discussions", params)
 
     # /changelog
     # ----------
 
-    @request
+    @request(scope=None)
     def changelog_build(self,
         stream: str,
         build: str
@@ -578,7 +846,7 @@ class OssapiV2:
         """
         return self._get(Build, f"/changelog/{stream}/{build}")
 
-    @request
+    @request(scope=None)
     def changelog_listing(self,
         from_: Optional[str] = None,
         to: Optional[str] = None,
@@ -591,7 +859,7 @@ class OssapiV2:
         params = {"from": from_, "to": to, "max_id": max_id, "stream": stream}
         return self._get(ChangelogListing, "/changelog", params)
 
-    @request
+    @request(scope=None)
     def changelog_lookup(self,
         changelog: str,
         key: Optional[str] = None
@@ -603,10 +871,27 @@ class OssapiV2:
         return self._get(Build, f"/changelog/{changelog}", params)
 
 
+    # /chat
+    # -----
+
+    @request(Scope.CHAT_WRITE)
+    def create_pm(self,
+        user_id: UserIdT,
+        message: str,
+        is_action: Optional[bool] = False
+    ) -> CreatePMResponse:
+        """
+        https://osu.ppy.sh/docs/index.html#create-new-pm
+        """
+        data = {"target_id": user_id, "message": message,
+            "is_action": is_action}
+        return self._post(CreatePMResponse, "/chat/new", data=data)
+
+
     # /comments
     # ---------
 
-    @request
+    @request(Scope.PUBLIC)
     def comments(self,
         commentable_type: Optional[CommentableTypeT] = None,
         commentable_id: Optional[int] = None,
@@ -629,7 +914,7 @@ class OssapiV2:
             "parent_id": parent_id, "sort": sort}
         return self._get(CommentBundle, "/comments", params)
 
-    @request
+    @request(scope=None)
     def comment(self,
         comment_id: int
     ) -> CommentBundle:
@@ -642,7 +927,7 @@ class OssapiV2:
     # /forums
     # -------
 
-    @request
+    @request(Scope.PUBLIC)
     def forum_topic(self,
         topic_id: int,
         cursor: Optional[Cursor] = None,
@@ -665,7 +950,7 @@ class OssapiV2:
     # / ("home")
     # ----------
 
-    @request
+    @request(Scope.PUBLIC)
     def search(self,
         mode: Optional[SearchModeT] = None,
         query: Optional[str] = None,
@@ -681,7 +966,7 @@ class OssapiV2:
     # /me
     # ---
 
-    @request
+    @request(Scope.IDENTIFY)
     def get_me(self,
         mode: Optional[GameModeT] = None
     ):
@@ -694,7 +979,7 @@ class OssapiV2:
     # /rankings
     # ---------
 
-    @request
+    @request(Scope.PUBLIC)
     def ranking(self,
         mode: GameModeT,
         type_: RankingTypeT,
@@ -712,7 +997,7 @@ class OssapiV2:
         return self._get(Rankings, f"/rankings/{mode.value}/{type_.value}",
             params=params)
 
-    @request
+    @request(Scope.PUBLIC)
     def spotlights(self) -> List[Spotlight]:
         """
         https://osu.ppy.sh/docs/index.html#get-spotlights
@@ -726,7 +1011,7 @@ class OssapiV2:
 
     # TODO add test for this once I figure out values for room_id and
     # playlist_id that actually produce a response lol
-    @request
+    @request(Scope.PUBLIC)
     def multiplayer_scores(self,
         room_id: int,
         playlist_id: int,
@@ -745,9 +1030,9 @@ class OssapiV2:
     # /users
     # ------
 
-    @request
+    @request(Scope.PUBLIC)
     def user_kudosu(self,
-        user_id: int,
+        user_id: UserIdT,
         limit: Optional[int] = None,
         offset: Optional[int] = None
     ) -> List[KudosuHistory]:
@@ -758,9 +1043,9 @@ class OssapiV2:
         return self._get(List[KudosuHistory], f"/users/{user_id}/kudosu",
             params)
 
-    @request
+    @request(Scope.PUBLIC)
     def user_scores(self,
-        user_id: int,
+        user_id: UserIdT,
         type_: ScoreTypeT,
         include_fails: Optional[bool] = None,
         mode: Optional[GameModeT] = None,
@@ -775,9 +1060,9 @@ class OssapiV2:
         return self._get(List[Score], f"/users/{user_id}/scores/{type_.value}",
             params)
 
-    @request
+    @request(Scope.PUBLIC)
     def user_beatmaps(self,
-        user_id: int,
+        user_id: UserIdT,
         type_: UserBeatmapTypeT,
         limit: Optional[int] = None,
         offset: Optional[int] = None
@@ -791,12 +1076,12 @@ class OssapiV2:
         if type_ is UserBeatmapType.MOST_PLAYED:
             return_type = List[BeatmapPlaycount]
 
-        return self._get(return_type, f"/users/{user_id}/beatmapsets/"
-            f"{type_.value}", params)
+        return self._get(return_type,
+            f"/users/{user_id}/beatmapsets/{type_.value}", params)
 
-    @request
+    @request(Scope.PUBLIC)
     def user_recent_activity(self,
-        user_id: int,
+        user_id: UserIdT,
         limit: Optional[int] = None,
         offset: Optional[int] = None
     ) -> List[Event]:
@@ -807,9 +1092,9 @@ class OssapiV2:
         return self._get(List[_Event], f"/users/{user_id}/recent_activity/",
             params)
 
-    @request
+    @request(Scope.PUBLIC)
     def user(self,
-        user: Union[int, str],
+        user: Union[UserIdT, str],
         mode: Optional[GameModeT] = None,
         key: Optional[UserLookupKeyT] = None
     ) -> User:
@@ -817,13 +1102,14 @@ class OssapiV2:
         https://osu.ppy.sh/docs/index.html#get-user
         """
         params = {"key": key}
-        return self._get(User, f"/users/{user}/{mode.value if mode else ''}", params)
+        return self._get(User, f"/users/{user}/{mode.value if mode else ''}",
+            params)
 
 
     # /wiki
     # -----
 
-    @request
+    @request(scope=None)
     def wiki_page(self,
         locale: str,
         path: str
@@ -837,45 +1123,41 @@ class OssapiV2:
     # undocumented
     # ------------
 
-    @request
+    @request(Scope.PUBLIC)
     def score(self,
         mode: GameModeT,
         score_id: int
     ) -> Score:
         return self._get(Score, f"/scores/{mode.value}/{score_id}")
 
-    @request
+    @request(Scope.PUBLIC, requires_login=True)
     def download_score(self,
         mode: GameModeT,
         score_id: int
-    ) -> str:
+    ) -> Replay:
         r = self.session.get(f"{self.BASE_URL}/scores/{mode.value}/"
             f"{score_id}/download")
+        replay = osrparse.parse_replay(r.content)
+        return Replay(replay, self)
 
-        tempfile = NamedTemporaryFile(mode="wb", delete=False)
-        with tempfile as f:
-            f.write(r.content)
-
-        return tempfile.name
-
-    @request
-    def search_beatmaps(self,
+    @request(Scope.PUBLIC)
+    def search_beatmapsets(self,
         query: Optional[str] = None,
         cursor: Optional[Cursor] = None
-    ) -> BeatmapSearchResult:
+    ) -> BeatmapsetSearchResult:
         # Param key names are the same as https://osu.ppy.sh/beatmapsets,
         # so from eg https://osu.ppy.sh/beatmapsets?q=black&s=any we get that
         # the query uses ``q`` and the category uses ``s``.
         # TODO implement all possible queries, or wait for them to be
         # documented. Currently we only implement the most basic "query" option.
         params = {"cursor": cursor, "q": query}
-        return self._get(BeatmapSearchResult, "/beatmapsets/search/", params)
+        return self._get(BeatmapsetSearchResult, "/beatmapsets/search/", params)
 
-    @request
+    @request(Scope.PUBLIC)
     def beatmapsets_events(self,
         limit: Optional[int] = None,
         page: Optional[int] = None,
-        user_id: Optional[int] = None,
+        user_id: Optional[UserIdT] = None,
         types: Optional[List[BeatmapsetEventTypeT]] = None,
         min_date: Optional[datetime] = None,
         max_date: Optional[datetime] = None
@@ -897,4 +1179,4 @@ class OssapiV2:
 
     def revoke_token(self):
         self.session.delete(f"{self.BASE_URL}/oauth/tokens/current")
-        self.clear_authentication()
+        self.remove_token(self.token_key, self.token_directory)
