@@ -1,6 +1,7 @@
 from unittest import TestCase
 
-from ossapi import RankingType, BeatmapsetEventType
+from ossapi import (RankingType, BeatmapsetEventType, AccessDeniedError,
+    InsufficientScopeError)
 
 from tests import api
 
@@ -10,7 +11,7 @@ class TestBeatmapsetDiscussionPosts(TestCase):
 
 class TestUserRecentActivity(TestCase):
     def test_deserialize(self):
-        api.user_recent_activity(10690090)
+        api.user_recent_activity(12092800)
 
 class TestSpotlights(TestCase):
     def test_deserialize(self):
@@ -66,18 +67,21 @@ class TestComment(TestCase):
 
 class TestDownloadScore(TestCase):
     def test_deserialize(self):
-        api.download_score(mode="osu", score_id=2797309065)
+        # api instance is using client credentials which doesn't have access to
+        # downloading replays
+        self.assertRaises(AccessDeniedError,
+            lambda: api.download_score(mode="osu", score_id=2797309065))
 
 class TestSearchBeatmaps(TestCase):
     def test_deserialize(self):
-        api.search_beatmaps(query="the big black")
+        api.search_beatmapsets(query="the big black")
 
 class TestUser(TestCase):
     def test_deserialize(self):
-        api.user(10690090)
+        api.user(12092800)
 
     def test_key(self):
-        # make suure it automatically falls back to username if not specified
+        # make sure it automatically falls back to username if not specified
         api.user("tybug2")
         api.user("tybug2", key="username")
 
@@ -85,9 +89,9 @@ class TestUser(TestCase):
 
 class TestMe(TestCase):
     def test_deserialize(self):
-        # TODO: requires another scope to be passed to OssapiV2
-        # api.get_me()
-        pass
+        # requires an authorization code api for the identify scope, client
+        # credentials can only request the public scope
+        self.assertRaises(InsufficientScopeError, api.get_me)
 
 class TestWikiPage(TestCase):
     def test_deserialize(self):
@@ -112,3 +116,14 @@ class TestForumTopic(TestCase):
 class TestBeatmapsetDiscussionVotes(TestCase):
     def test_deserialize(self):
         api.beatmapset_discussion_votes().votes[0].score
+
+class TestBeatmapsetDiscussions(TestCase):
+    def test_deserialize(self):
+        api.beatmapset_discussions()
+
+# TODO revisit when either ``api`` is scoped differently (requires redirect_uri
+# stored) or auth handling is rewriting to allow multiple clients
+# class TestCreateNewPM(TestCase):
+#     def test_deserialize(self):
+#         api.create_pm(2070907, "Unit test from ossapi "
+#             "(https://github.com/circleguard/ossapi/), please ignore")
